﻿/*
 * Készítette a SharpDevelop.
 * Felhasználó: phil
 * Dátum: 2009.06.22.
 * Idő: 15:40
 * 
 * A sablon megváltoztatásához használja az Eszközök | Beállítások | Kódolás | Szabvány Fejlécek Szerkesztését.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.IO.Ports;
using System.Text;
using System.Threading;

namespace Noise.DataAcquisition
{
	[Flags]
	public enum ResistanceSwitching : byte {
		All = 15,
		Channel1 = 1,
		Channel2 = 2,
		Channel3 = 4,
		Channel4 = 8,
		None = 0,
		SupplyVoltage = 16,
		NotDefined = 32
	}
	
	[Flags]
	public enum Adc : byte {
		Adc1 = 1,
		Adc2 = 2,
		Both = 3,
		None = 0,
		NotDefined = 8
	}
	
	public enum Dac : byte {
		Dac1,
		Dac2
	}
	
	public class EduDaq : ISerialDataCollector
	{
		#region CONSTANTS
		public const string AdcChoiceCommand = "@A";
		public const string CommandToken = "@";
		public const string DeviceInfoCommand = "@I";
		public const string MeasureCommand = "@M";
		public const string MeasureSeriesCommand = "@s";
		public const string ResistanceSwitchingCommand = "@x";
		public const string SamplingFrequencyCommand = "@f";
		public const string SetAdc1Command = "@1";
		public const string SetAdc2Command = "@2";
		public const string SetDac1Command = "@d";
		public const string SetDac2Command = "@D";
		public const string StartSamplingCommand = "@S";
		public const string TriggerInvertCommand = "@t";
		public static readonly string[] ChannelNames = new string[] { "A", "B", "C", "D" };
		public const int BlockLength = 2;	// The number of data points (not bytes!) transmitted in one transaction. TODO: This is obsolete.
		public const int WordLength = 2; 	// The number of bytes per data point.
		#endregion
		#region FIELDS
		protected bool? triggerInverted = null;
		protected double? samplingFrequency = null;
		protected EventWaitHandle ready = new ManualResetEvent(true);
		protected EventWaitHandle samplingStarted = new ManualResetEvent(false);
		protected EventWaitHandle samplingStopped = new ManualResetEvent(true);
		protected Adc activeAdc = Adc.Both;
		protected IChannelDataSocket<double[]> dataSocket; //TODO: this is to be replaced by a List<IChannelDataSocket<double>> channels;
		protected ISerialPort port;
		protected ResistanceSwitching resistanceSwitching = ResistanceSwitching.NotDefined;
		protected AdcInfo[] adcStates = new AdcInfo[2];	// TODO: Remove this.
		protected double?[] dacStates = new Nullable<double>[] { null, null };
		protected BackgroundWorker worker;
		#endregion
		#region EVENTS
		public event AsyncCompletedEventHandler Stopped;
		#endregion
		
		#region CONSTRUCTORS
		public EduDaq(ISerialPort port)
		{
			this.port = port;
		}	
		#endregion
		
		#region PROPERTIES
		public Adc ActiveAdc {
			get { return activeAdc; }
		}
		
		public AdcInfo[] AdcStates {
			get { return adcStates; }
		}
		
		public Nullable<double>[] DacStates {
			get { return dacStates; }
		}
			
		public string Name {
			get { return "EDaq"; }
		}
		
		public IChannelDataSocket<double[]> DataSocket {
			get { return dataSocket; }
			set { dataSocket = value; }
		}
		
		public ISerialPort Port {
			get {
				lock (port.SyncRoot) {
					return port;
				}		
			}
		}
		
		public ResistanceSwitching ResistanceSwitching {
			get { return resistanceSwitching; }
		}
		
		public Nullable<double> SamplingFrequency {
			get { return samplingFrequency; }
		}
		
		public EventWaitHandle SamplingStarted {
			get { return samplingStarted; }
		}
		
		public EventWaitHandle SamplingStopped {
			get { return samplingStopped; }
		}
		
		public Nullable<bool> TriggerInverted {
			get { return triggerInverted; }
		}
		#endregion
		
		#region PUBLIC METHODS
		public void Escape()
		{
			Port.Write(new byte[] { 27 });
			Thread.Sleep(600);
		}
		
		public string GetDeviceInfo()
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			string deviceInfo = String.Empty;
			byte[] commandBytes = ConvertToBytes(DeviceInfoCommand, null);
			Encoding ascii = Encoding.ASCII;
			bool continueReading = true;
			string command = "§";
			byte[] cb = ConvertToBytes(command, null);
			byte[] read;
			//
			Send(commandBytes);
			while (continueReading) {
				Port.Write(cb);
				read = Port.Read(1);
				continueReading = (read[0] != cb[0]);
				if (continueReading) {
					deviceInfo += ascii.GetString(read);
				}
			}
			//
			ready.Set();
			return deviceInfo;
		}
		
		public double[] Measure(byte averages)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			byte[] commandBytes = ConvertToBytes(MeasureCommand, new byte[] { averages > 0 ? averages : (byte)1 });
			byte[] dataBytes;
			double[] measuredVoltages;
			int blockByteCount = BlockLength * WordLength;
			//
			Send(commandBytes);
			dataBytes = Port.Read(blockByteCount);
			measuredVoltages = ConvertToVoltages(dataBytes);
			//
			ready.Set();
			return measuredVoltages;
		}
		
		public double[] MeasureSeries(int samples, double frequency)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			byte[] samplesBytes = BitConverter.GetBytes((ushort)samples);
			byte[] paramBytes = new byte[5];
			byte[] freqBytes = BitConverter.GetBytes((int)frequency);
			byte[] commandBytes;
			byte[] dataBytes;
			double[] measuredVoltages;
			int blockByteCount = (activeAdc == Adc.Both) ? BlockLength * WordLength : WordLength;
			//
			paramBytes[0] = BitConverter.IsLittleEndian ? freqBytes[2] : freqBytes[1];
			paramBytes[1] = BitConverter.IsLittleEndian ? freqBytes[1] : freqBytes[2];
			paramBytes[2] = BitConverter.IsLittleEndian ? freqBytes[0] : freqBytes[3];
			paramBytes[3] = BitConverter.IsLittleEndian ? samplesBytes[1] : samplesBytes[0];
			paramBytes[4] = BitConverter.IsLittleEndian ? samplesBytes[0] : samplesBytes[1];
			commandBytes = ConvertToBytes(MeasureSeriesCommand, paramBytes);
			try {
				Send(commandBytes);
				dataBytes = port.Read(blockByteCount * samples);
				measuredVoltages = ConvertToVoltages(dataBytes);
			} catch (Exception exception) {
				measuredVoltages = null;
				throw exception;
			} finally {
				//
				ready.Set();
			}
			return measuredVoltages;
		}
		
		public void Reset()
		{
			if ((Port != null) && Port.IsOpen) {
				ready.Reset();
				//
				Escape();
				Port.DiscardOutBuffer();
				Port.DiscardInBuffer();
				//
				ready.Set();
			} else {
				ready.Reset();
			}
		}
		
		public void Send(string command, byte[] parameters)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			Send(ConvertToBytes(command, parameters));
			//
			ready.Set();
		}
		
		public void Send(string command)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			Send(ConvertToBytes(command, null));
			//
			ready.Set();
		}
		
		public void SetAdc(Adc adc, string channel, double gain)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			int adcStateIndex;
			int gainExponent = (int)Math.Round(Math.Log(gain) / Math.Log(2), MidpointRounding.AwayFromZero);
			if (gainExponent > 15) gainExponent = 15;
			double setGain = Math.Pow(2.0, gainExponent);
			byte parameter;
			byte gainByte = (byte)gainExponent;
			string commandString;
			string setChannel;
			AdcInfo adcInfo = null;
			gainByte <<= 4;
			
			switch (adc) {
				case Adc.Adc2:
					adcStateIndex = 1;
					if (channel.ToUpper() == "D") {
						parameter = (byte)1;
						setChannel = "D";
					} else {
						parameter = (byte)0;
						setChannel = "C";
					}
					commandString = SetAdc2Command;
					break;
				default:
				case Adc.Adc1:
					adcStateIndex = 0;
					if (channel.ToUpper() == "B") {
						parameter = (byte)1;
						setChannel = "B";
					} else {
						parameter = (byte)0;
						setChannel = "A";
					}
					commandString = SetAdc1Command;
					break;
			}
			
			parameter |= gainByte;
			
			try {
				Send(commandString, new byte[] { parameter });
				adcInfo = new AdcInfo(setChannel, setGain);
			} catch (Exception exception) {
				adcInfo = null;
				throw exception;
			} finally {
				adcStates[adcStateIndex] = adcInfo;
				//
				ready.Set();
			}
		}
		
		public void SetActiveAdc(Adc adc)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			byte[] commandBytes = ConvertToBytes(AdcChoiceCommand, new byte[] { (byte)adc });
			//
			try {
				Send(commandBytes);
				activeAdc = adc;
			} catch (Exception exception) {
				activeAdc = Adc.NotDefined;
				throw exception;
			}
			//
			ready.Set();
		}
		
		public void SetDac(Dac dac, double voltage)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			int dacStateIndex;
			ushort z = (ushort)(32768 * (voltage / 5.0 + 1));
			double setVoltage = 5.0 * (z / 32768.0  - 1.0);
			byte[] zBytes = BitConverter.GetBytes(z);
			byte[] paramBytes = new byte[2];
			byte[] commandBytes;
			
			paramBytes[0] = BitConverter.IsLittleEndian ? zBytes[1] : zBytes[0];
			paramBytes[1] = BitConverter.IsLittleEndian ? zBytes[0] : zBytes[1];
			
			switch (dac) {
				case Dac.Dac2:
					commandBytes = ConvertToBytes(SetDac2Command, paramBytes);
					dacStateIndex = 1;
					break;
				case Dac.Dac1:
				default:
					commandBytes = ConvertToBytes(SetDac1Command, paramBytes);
					dacStateIndex = 0;
					break;
			}
			
			try {
				Send(commandBytes);
				dacStates[dacStateIndex] = setVoltage;
			} catch (Exception exception) {
				dacStates[dacStateIndex] = null;
				throw exception;
			} finally {
				//
				ready.Set();
			}		
		}
		
		public void SetResistanceSwitching(ResistanceSwitching switching)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			byte[] commandbytes = ConvertToBytes(ResistanceSwitchingCommand, new byte[] { (byte)switching });
			//
			try {
				Send(commandbytes);
				resistanceSwitching = switching;
			} catch (Exception exception) {
				resistanceSwitching = ResistanceSwitching.NotDefined;
				throw exception;
			}
			//
			ready.Set();
		}
		
		public void SetSamplingFrequency(double samplingFrequency)
		{
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();
			//
			ushort f = (ushort)samplingFrequency;
			byte[] paramBytes = new byte[2];
			byte[] commandBytes;
			int remainder;
			//
			paramBytes[0] = (byte)Math.DivRem(f, 256, out remainder);
			paramBytes[1] = (byte)remainder;
			commandBytes = ConvertToBytes(SamplingFrequencyCommand, paramBytes);
			try {
				Send(commandBytes);
				this.samplingFrequency = samplingFrequency;
			} catch (Exception exception) {
				this.samplingFrequency = null;
				throw exception;
			}
			//
			ready.Set();
		}
		
		public void SetTriggerInversion(bool invert)
		{
			try {
				Send(ConvertToBytes(TriggerInvertCommand, new byte[] { invert ? (byte)1 : (byte)0 }));
				triggerInverted = invert;
			} catch (Exception exception) {
				triggerInverted = null;
				throw exception;
			}
		}
		
		public void StartSampling(double samplingFrequency)
		{
			SetSamplingFrequency(samplingFrequency);
			StartSampling();
		}
		
		public void StartSampling()
		{
			if (!samplingFrequency.HasValue) {
				Exception exception = new InvalidOperationException("Sampling frequency is not set.");
				throw exception;
			}
			if (dataSocket == null) {
				Exception exception = new InvalidOperationException("No data socket defined.");
				throw exception;
			}
			CheckContinuousMode();
			CheckIfReady();
			ready.Reset();	// Will be set after the BackgroundWorker completes work.
			//
			byte[] commandBytes = ConvertToBytes(StartSamplingCommand, null);
			//
			Send(commandBytes);
			worker = new BackgroundWorker();
			worker.WorkerSupportsCancellation = true;
			//worker.DoWork += Sample;
			worker.DoWork += SampleInChunks;
			worker.RunWorkerCompleted += HandleRunWorkerCompletedEvents;
			worker.RunWorkerAsync();
			samplingStarted.Set();
			samplingStopped.Reset();
		}
		
		public void StopSampling()
		{
			if ((worker != null) && worker.IsBusy) {
				worker.CancelAsync();
			}
		}
		#endregion
		
		#region PROTECTED METHODS
		protected void CheckContinuousMode()
		{
			if (samplingStarted.WaitOne(0)) {
				Exception exception = new InvalidOperationException("Operation cannot be executed whilst the device is in continuous mode.");
				throw exception;
			}
		}
		
		protected void CheckIfReady()
		{
			if (!ready.WaitOne(1000)) {
				Exception exception = new InvalidOperationException("Timeout in waiting for the device to be ready.");
				throw exception;
			}
		}
			
		protected byte[] ConvertToBytes(string command, byte[] parameters)
		{
			byte[] commandBytes = Encoding.ASCII.GetBytes(command);
			byte[] bytesToSend = (parameters != null) ? new byte[commandBytes.Length + parameters.Length] : new byte[commandBytes.Length];
			Array.Copy(commandBytes, bytesToSend, commandBytes.Length);
			if (parameters != null) {
				Array.Copy(parameters, 0,  bytesToSend, commandBytes.Length, parameters.Length);
			}
			
			return bytesToSend;
		}
		
		protected double[] ConvertToVoltages(byte[] bytes)
		{
			double[] voltages = new double[bytes.Length / WordLength];
			int v;	// Voltage as 16-bit code.
			
			for (int i = 0; i < voltages.Length; i++) {
				v = bytes[WordLength * i] << 8;	// TODO: This code assumes 2-byte words, will not work for other word lengths.
				v += bytes[WordLength * i + 1];
				voltages[i] = 5.0 * (v / 32768.0  - 1.0);	// 16 bits between -5 V and +5 V; this is [10.0 * v / 65536.0 - 5.0] simplified.
			}
			
			return voltages;
		}
		
		protected void HandleRunWorkerCompletedEvents(object sender, RunWorkerCompletedEventArgs e)
		{
			Reset();	// Also sets 'ready'.
			AsyncCompletedEventArgs completedArgs = new AsyncCompletedEventArgs(e.Error, true, null);
			OnStopped(completedArgs);
		}
		
		protected virtual void OnStopped(AsyncCompletedEventArgs e)
		{
			if (Stopped != null) {
				Stopped(this, e);
			}
		}
		
		protected void Sample(object sender, DoWorkEventArgs e)
		{
			BackgroundWorker theWorker = sender as BackgroundWorker;
			byte[] read = null;
			double[] voltages;
			int blockByteCount = (activeAdc == Adc.Both) ? BlockLength * WordLength : WordLength;
			bool dataAvailable = false;
			bool continueCleaning = true;
			int idleTime = 2;
			int idleCounter = 0;
			int cleaningTimeOutCounter = 2000;
			
			try {				
				while (!theWorker.CancellationPending) {
					// We do not use the Port property to avoid overhead, so every attempt at accessing the port is within a lock.
					lock (port.SyncRoot) {
						if (port.BytesToRead >= blockByteCount) {
							read = port.Read(blockByteCount);
							dataAvailable = true;
						} else {
							dataAvailable = false;
						}
					}
					if (dataAvailable) {
						voltages = ConvertToVoltages(read);
						lock ((dataSocket as ICollection).SyncRoot) {
							dataSocket.Enqueue(voltages);
						}
					} else {
						Thread.Sleep(idleTime);
					}
				}
				// Notify the device to exit continuous sampling mode.
				Escape();
				// Read the remaining data:
				while (continueCleaning) {
					// We do not use the Port property to avoid overhead, so every attempt at accessing the port is within a lock.
					lock (port.SyncRoot) {
						if (port.BytesToRead >= blockByteCount) {
							read = port.Read(blockByteCount);
							dataAvailable = true;
						} else {
							dataAvailable = false;
						}
					}
					if (dataAvailable) {
						idleCounter = 0;
						voltages = ConvertToVoltages(read);
						lock ((dataSocket as ICollection).SyncRoot) {
							dataSocket.Enqueue(voltages);
						}
					} else {
						idleCounter++;
						Thread.Sleep(idleTime);
						continueCleaning = (idleCounter < cleaningTimeOutCounter);
					}
				}
			} catch (Exception exception) {
				Escape();
				exception.Data.Add("Bytes at port", port.BytesToRead);
				throw exception;
			} finally {
				SignalStop();
			}
		}
		
		// Test method to see if it improves performance.
		protected void SampleInChunks(object sender, DoWorkEventArgs e)
		{
			BackgroundWorker theWorker = sender as BackgroundWorker;
			byte[] read = null;
			byte[] block;
			double[] voltages;
			int blockByteCount = (activeAdc == Adc.Both) ? BlockLength * WordLength : WordLength;
			int bytesAtPort = 0;
			int blocksAtPort = 0;
			int idleTime = 2;
			int idleCounter = 0;
			int cleaningTimeOutCounter = 200;
			bool continueCleaning = true;
			
			try {				
				while (!theWorker.CancellationPending) {
					// We do not use the Port property to avoid overhead, so every attempt at accessing the port is within a lock.
					lock (port.SyncRoot) {
						bytesAtPort = port.BytesToRead;
						blocksAtPort = bytesAtPort / blockByteCount;
						if (blocksAtPort > 0) {
							read = port.Read(blocksAtPort * blockByteCount);
						} else {
							read = null;
						}
					}
					if ((read != null) && (read.Length > 0)) {
						for (int i = 0; i < read.Length - blockByteCount; i += blockByteCount) {
							block = new byte[blockByteCount];
							for (int j = 0; j < block.Length; j++) {
								block[j] = read[i + j];
							}
							voltages = ConvertToVoltages(block);
							lock (dataSocket.SyncRoot) {
								dataSocket.Enqueue(voltages);
							}
						}
					} else {
						Thread.Sleep(idleTime);
					}
				}
				// Notify the device to exit continuous sampling mode.
				Escape();
				// Read the remaining data:
				while (continueCleaning) {
					// We do not use the Port property to avoid overhead, so every attempt at accessing the port is within a lock.
					lock (port.SyncRoot) {
						bytesAtPort = port.BytesToRead;
						blocksAtPort = bytesAtPort / blockByteCount;
						if (blocksAtPort > 0) {
							read = port.Read(blocksAtPort * blockByteCount);
						} else {
							read = null;
						}
					}
					if ((read != null) && (read.Length > 0)) {
						idleCounter = 0;
						for (int i = 0; i < read.Length - blockByteCount; i += blockByteCount) {
							block = new byte[blockByteCount];
							for (int j = 0; j < block.Length; j++) {
								block[j] = read[i + j];
							}
							voltages = ConvertToVoltages(block);
							lock (dataSocket.SyncRoot) {
								dataSocket.Enqueue(voltages);
							}
						}
					} else {
						idleCounter++;
						Thread.Sleep(idleTime);
						continueCleaning = (idleCounter < cleaningTimeOutCounter);
					}
				}
			} catch (Exception exception) {
				Escape();
				exception.Data.Add("Bytes at port", port.BytesToRead);
				throw exception;
			} finally {
				SignalStop();
			}
		}
		
		protected void Send(byte[] bytes)
		{
			byte[] read;
			foreach (byte b in bytes) {
				Port.Write(new byte[] { b });
				read = Port.Read(1);
				if (read[0] != b) {
					Exception exception = new IOException("Sent and received bytes do not match.");
					exception.Data.Add("Sent byte", b);
					exception.Data.Add("Received byte", read[0]);
					throw exception;
				}
			}
		}
		
		protected void SignalStop()
		{
			samplingStarted.Reset();
			samplingStopped.Set();
		}
		#endregion
	}
}
