/*
 * Created by SharpDevelop.
 * User: phil
 * Date: 2010.02.19.
 * Time: 15:42
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Xml;
//
using TetheredSun.Core;

namespace TetheredSun.SignalAnalysis
{
	/// <summary>
	/// Description of LevelCrossingDetector.
	/// </summary>
	public class LevelCrossingDetector : IXmlConvertible
	{
		protected object syncRoot = new Object();
		protected double hysteresis = 0.0;
		protected double level;
		protected double oldTime = 0.0;				// The time of a previous crossing candidate.
		protected double oldValue = Double.NaN;		// The value at a previous crossing candidate.
		protected double lastTime = Double.NaN;		// The previous time instance, regardless of crossing.
		protected double lastValue = Double.NaN;	// The previous value tested, regardless of crossing.
		protected volatile bool isActive = true;
		protected volatile bool isUpCrossing = false;
		protected volatile bool isDownCrossing = false;
		protected bool downBegun = false;
		protected bool upBegun = false;
		protected LevelCrossingCycleCollection cycles = new LevelCrossingCycleCollection();
		Point<double> crossingPoint;
		
		public event EventHandler CyclesUpdated;
		
		
		public LevelCrossingDetector(double level, double hysteresis)
		{
			this.level = level;
			this.hysteresis = hysteresis;
		}
		
		public LevelCrossingDetector(double level) : this(level, 0.0) { }
		
		public Point<double> CrossingPoint {
			get { return crossingPoint; }
		}
		
		public LevelCrossingCycleCollection Cycles {
			get { lock (cycles.SyncRoot) { return cycles; } }
		}
		
		public double Hysteresis {
			get { lock (syncRoot) { return hysteresis; } }
			set { lock (syncRoot) { hysteresis = value; } }
		}
		
		public bool IsActive {
			get { return isActive; }
			set { isActive = value; }
		}
		
		public bool IsUpCrossing {
			get { return isUpCrossing; }
		}
		
		public bool IsDownCrossing {
			get { return isDownCrossing; }
		}
		
		public double Level {
			get { lock (syncRoot) { return level; } }
			set { lock (syncRoot) { level = value; } }
		}
	
		
		public void FromXml(XmlElement xmlElement)
		{
			if (xmlElement.Name != "LevelCrossingDetector") {
				Exception exception = new ArgumentException("Invalid element name.");
				
				exception.Data.Add("Expected name", "LevelCrossingDetector");
				exception.Data.Add("Received name", xmlElement.Name);
				
				throw exception;
			}
			
			XmlNode child;
			double parsedDouble;
			bool parsedBool;
			
			child = xmlElement.SelectSingleNode("child::Level");
			level = Double.TryParse(child.InnerXml, out parsedDouble) ? parsedDouble : 1.0;
			
			child = xmlElement.SelectSingleNode("child::Hysteresis");
			hysteresis = Double.TryParse(child.InnerXml, out parsedDouble) ? parsedDouble : 0.0;
			
			child = xmlElement.SelectSingleNode("child::IsActive");
			isActive = Boolean.TryParse(child.InnerXml, out parsedBool) ? parsedBool : true;
		}
		
		public XmlElement ToXml(XmlDocument context)
		{
			XmlElement detectorXml = context.CreateElement("LevelCrossingDetector");
			XmlElement levelXml = context.CreateElement("Level");
			XmlElement hysteresisXml = context.CreateElement("Hysteresis");
			XmlElement isActiveXml = context.CreateElement("IsActive");
			
			levelXml.InnerXml = level.ToString();
			detectorXml.AppendChild(levelXml);
			
			hysteresisXml.InnerXml = hysteresis.ToString();
			detectorXml.AppendChild(hysteresisXml);
			
			isActiveXml.InnerXml = isActive.ToString();
			detectorXml.AppendChild(isActiveXml);
			
			return detectorXml;
		}
		
		
		public void Test(double time, double value)
		{
			if (!isActive) return;
			if (Double.IsNaN(lastTime) || Double.IsNaN(lastValue)) {
				lastTime = time;
				lastValue = value;
				return;
			}
			
			double level1;
			double level2;
			isUpCrossing = false;
			isDownCrossing = false;
			
			lock (syncRoot) {
				level1 = level - hysteresis;
				level2 = level + hysteresis;
			}
			
			if (value > level2) {
				if (lastValue <= level1) {	// We have an immediate upwards crossing.
					// Clear crossing start indicators to be sure.
					upBegun = false;
					downBegun = false;
					oldTime = lastTime;
					oldValue = lastValue;
					AddNewCycle(GetCrossing(time, value));
				} else if (lastValue <= level2) {
					if (upBegun) {		// Complete the upwards level-crossing candidate.
						upBegun = false;
						downBegun = false;
						AddNewCycle(GetCrossing(time, value));
					} else if (downBegun) {	// Cancel the downwards level-crossing candidate.
						downBegun = false;
					}
				}
			} else if ((value < level2) && (value > level1)) {
				if (lastValue >= level2) {	// We have a downwards level-crossing candidate.
					upBegun = false;
					downBegun = true;
					oldTime = time;
					oldValue = value;
				} else if (lastValue <= level1) {
					upBegun = true;
					downBegun = false;
					oldTime = time;
					oldValue = value;
				}
			} else if (value < level1) {
				if (lastValue >= level2) {	// We have an immediate downwards crossing.
					// Clear crossing start indicators to be sure.
					upBegun = false;
					downBegun = false;
					oldTime = lastTime;
					oldValue = lastValue;
					CompleteCycle(GetCrossing(time, value));
				} else if (lastValue >= level1) {
					if (downBegun) {
						upBegun = false;
						downBegun = false;
						CompleteCycle(GetCrossing(time, value));
					} else if (upBegun) {
						upBegun = false;
					}
				}
			}
			
			lastTime = time;
			lastValue = value;
		}
		
		protected void AddNewCycle(double upCrossingTime)
		{
			// Create a new cycle:
			isUpCrossing = true;
			isDownCrossing = false;
			// A new LevelCrossingCycle must be added
			// (note: we lock on the SyncRoot of cycle):
			Cycles.Add(new LevelCrossingCycle(upCrossingTime));	// This also takes care of the nextUpCrossing
			// of the previous item.
			OnCyclesUpdated(new EventArgs());
		}
		
		protected void CompleteCycle(double downCrossingTime)
		{
			isDownCrossing = true;
			isUpCrossing = false;
			lock (cycles.SyncRoot) {
				int count = cycles.Count;
				if (count > 0) {
					cycles[count - 1].DownCrossing = downCrossingTime;
				}
			}
			OnCyclesUpdated(new EventArgs());
		}
		
		protected double GetCrossing(double time, double value)
		{
			double slope = (value - oldValue) / (time - oldTime);
			double t = oldTime + (level - oldValue) / slope;
			double x = oldValue + (t - oldTime) * slope;
			
			crossingPoint = new Point<double>(t, x);
			
			return t;
		}
		
		protected virtual void OnCyclesUpdated(EventArgs e)
		{
			if (CyclesUpdated != null) {
				CyclesUpdated(this, e);
			}
		}
		
	}
}
